from pyrbit.mem_utils import MemoryModel
from pyrbit.mle_utils import mle_sequence
from pyrbit.plot_utils import regplot
import seaborn
import numpy
import matplotlib.pyplot as plt


class ACTR(MemoryModel):
    def __init__(self, nitems, d, s, tau, buffer_size=256, seed=None):
        super().__init__(nitems, seed=seed)
        self.d, self.s, self.tau = d, s, tau
        self.buffer_size = buffer_size
        self.reset()

    def reset(self, d=None, s=None, tau=None):
        if d is not None:
            self.d = d
        if s is not None:
            self.s = s
        if tau is not None:
            self.tau = tau

        self.counter_pres = numpy.zeros((self.nitems,))
        self.counter_time = numpy.full((self.nitems, self.buffer_size), numpy.inf)

    def _print_info(self):
        print(f"d: \t {self.counter_pres}")
        print(f"time: \t {self.counter_time}")

    def _compute_item_activation(self, item, time=None):
        if time is None:
            time = numpy.max(self.counter_time)
        self.activation = numpy.log(
            numpy.sum((time - self.counter_time[item]) ** (-self.d))
        )
        return self.activation

    def _compute_item_probabilities(self, item, time=None):
        activation = self._compute_item_activation(item, time=time)
        return 1 / (1 + numpy.exp((self.tau - activation) / self.s))

    def compute_probabilities(self, time=None):
        cip = numpy.vectorize(self._compute_item_probabilities)
        return cip(list(range(self.nitems)), time=time)

    def update(self, item, time):
        if len(time) == 1:
            n = numpy.asarray(self.counter_pres[item], dtype=numpy.int32)
            self.counter_time[item, n] = time
            self.counter_pres[item] += 1
        else:
            n = numpy.asarray(self.counter_pres[item], dtype=numpy.int32)
            self.counter_time[item, n : n + len(time)] = time
            self.counter_pres[item] += len(time)


# helper functions for simulations
def simulate_arbitrary_traj(actr, N, seed=None):
    recall = []
    times = []
    query_times = []
    rng = numpy.random.default_rng(seed=seed)

    for trial in range(N):
        actr.reset()
        nrepet = rng.integers(low=1, high=10, size=(1,))
        _times = numpy.sort(rng.random(size=(int(nrepet),)) * 100)
        times.append(_times)
        actr.update(0, _times)
        query_time = rng.random(size=(1,)) * 100 + _times[-1]
        query_times.append(query_time)
        recall.append(actr.query_item(0, query_time))
    return recall, times, query_times


def gen_data(actr_model, N, seed=None):
    recalls, times, query_times = simulate_arbitrary_traj(actr_model, N, seed=seed)
    recalls = [int(r[0]) for r in recalls]
    deltatis = [qtime - time for time, qtime in zip(times, query_times)]
    return recalls, deltatis


## visualisations


def diagnostics(d, deltatis, recall, ax=None, **kwargs):
    return logregplot(d, deltatis, recall, ax=None, **kwargs)


def logregplot(d, deltatis, recall, ax=None, label_prefix="", **kwargs):
    x_bins = kwargs.pop("x_bins", None)
    if x_bins is None:
        x_bins = int(len(recall) ** (1 / 3))
    activation = [
        numpy.log(numpy.sum(numpy.asarray(deltati) ** -d)) for deltati in deltatis
    ]
    return _logregplot_activation(
        activation, recall, ax=ax, x_bins=x_bins, label_prefix=label_prefix, **kwargs
    )


## inference


def identify_actr_from_recall_sequence(
    recall_sequence,
    deltas,
    guess=(0.5, 0.25, -0.7),
    optim_kwargs={"method": "L-BFGS-B", "bounds": [(0, 1), (-5, 5), (-5, 5)]},
    verbose=True,
    basin_hopping=False,
    basin_hopping_kwargs=None,
):
    infer_results = mle_sequence(
        _actr_sequence_likelihood_transform,
        optim_kwargs,
        guess,
        deltas,
        recall_sequence,
        basin_hopping=basin_hopping,
        basin_hopping_kwargs=basin_hopping_kwargs,
    )

    if verbose:
        print(infer_results)

    return infer_results


def actr_observed_information_matrix(recall_sequence, deltatis, d, s, tau):
    return _actr_get_sequence_observed_information_matrix(
        recall_sequence, deltatis, d, s, tau
    )[0]


def _actr_get_sequence_observed_information_matrix(
    recall_sequence, deltatis, d, s, tau
):
    J_11 = 0
    J_12 = 0
    J_13 = 0
    J_22 = 0
    J_23 = 0
    J_33 = 0

    for recall, deltati in zip(
        recall_sequence,
        deltatis,
    ):
        if recall == 1:
            J_11 += actr_ddq1_dd_dd_sample(tau, s, d, deltati)
            J_12 += actr_ddq1_dd_ds_sample(tau, s, d, deltati)
            J_13 += actr_ddq1_dd_dtau_sample(tau, s, d, deltati)
            J_22 += actr_ddq1_ds_ds_sample(tau, s, d, deltati)
            J_23 += actr_ddq1_ds_dtau_sample(tau, s, d, deltati)
            J_33 += actr_ddq1_dtau_dtau_sample(tau, s, d, deltati)
        elif recall == 0:
            J_11 += actr_ddq0_dd_dd_sample(tau, s, d, deltati)
            J_12 += actr_ddq0_dd_ds_sample(tau, s, d, deltati)
            J_13 += actr_ddq0_dd_dtau_sample(tau, s, d, deltati)
            J_22 += actr_ddq0_ds_ds_sample(tau, s, d, deltati)
            J_23 += actr_ddq0_ds_dtau_sample(tau, s, d, deltati)
            J_33 += actr_ddq0_dtau_dtau_sample(tau, s, d, deltati)
        else:
            raise ValueError(f"recall is not either 1 or 0, but is {recall}")

    J = -numpy.array([[J_11, J_12, J_13], [J_12, J_22, J_23], [J_13, J_23, J_33]])
    return J, len(recall_sequence)


def _actr_sequence_likelihood_transform(
    theta,
    deltas,
    recalls,
    transform=None,
):
    ll = 0
    d, s, tau = theta
    if transform is None:
        transform = lambda d, s, tau: (d, s, tau)
    for n, (deltati, recall) in enumerate(zip(deltas, recalls)):
        dll = _actr_log_likelihood_sample(recall, deltati, d, s, tau, transform)
        ll += dll

    return ll


def _logregplot_activation(activation, recall, ax=None, label_prefix="", **kwargs):
    ci = kwargs.pop("ci", None)
    recall_event_kwargs = kwargs.pop("recall_event_kwargs", {})

    if ax is None:
        fig, ax = plt.subplots(nrows=1, ncols=1)

    regplot(
        x=activation,
        y=recall,
        scatter=False,
        fit_reg=True,
        logistic=True,
        ax=ax,
        ci=ci,
        label=label_prefix + "Logistic Regression",
        **kwargs,
    )

    regplot(
        x=activation,
        y=recall,
        scatter=True,
        fit_reg=False,
        ax=ax,
        label=label_prefix + "Recall events",
        **recall_event_kwargs,
    )

    regplot(
        x=activation,
        y=recall,
        scatter=True,
        fit_reg=False,
        ax=ax,
        label=label_prefix + "Estimated recall probability",
        **kwargs,
    )

    ax.set_ylabel("Recall probability")
    ax.set_xlabel("Activation")
    return ax


def _actr_log_likelihood_sample(recall, deltas, d, s, tau, transform):
    # rescaling value to linear
    d, s, tau = transform(d, s, tau)
    if recall == 1:  # Warning: passing to array converts recall to float
        return actr_q1_sample(tau, s, d, deltas)
    elif recall == 0:
        return actr_q0_sample(tau, s, d, deltas)

    else:
        raise ValueError(f"Recall is not 0 or 1, but is {recall}")


def logistic(x):
    return 1 / (1 + numpy.exp(-x))


def activation(deltati, d):
    deltati = numpy.asarray(deltati)
    return numpy.log(numpy.sum(deltati ** (-d)))


def softplus(x):
    return numpy.log(1 + numpy.exp(x))


## ============ for observed information matrix ============= ##
## p1, q1, p0, q0
def actr_p1_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return logistic(X)


def actr_p0_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return logistic(-X)


def actr_q1_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -softplus(-X)


def actr_q0_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -softplus(X)


## first order derivatives


def da_dd(deltati, d):
    deltati = numpy.asarray(deltati)
    num = -numpy.sum(numpy.log(deltati) * deltati ** (-d))
    denom = numpy.sum(deltati ** (-d))
    return num / denom


def actr_dq1_ds_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return (tau - a) / s**2 * logistic(-X)


def actr_dq1_dtau_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -1 / s * logistic(-X)


def actr_dq1_dd_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return 1 / s * logistic(-X) * da_dd(deltati, d)


def actr_dq0_ds_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -(tau - a) / s**2 * logistic(X)


def actr_dq0_dtau_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return 1 / s * logistic(X)


def actr_dq0_dd_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -1 / s * logistic(X) * da_dd(deltati, d)


## Second order derivatives


def dda_ddd(deltati, d):
    deltati = numpy.asarray(deltati)
    num1 = numpy.sum(numpy.log(deltati) ** 2 * deltati ** (-d)) * numpy.sum(
        deltati ** (-d)
    )
    num2 = -((numpy.sum(numpy.log(deltati) * deltati ** (-d))) ** 2)
    denom = (numpy.sum(deltati ** (-(d)))) ** 2
    return (num1 + num2) / denom


# omega = 1
def actr_ddq1_ds_ds_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return 2 * (a - tau) / s**3 * logistic(-X) - (tau - a) ** 2 / s**4 * (
        logistic(X) * logistic(-X)
    )


def actr_ddq1_dtau_dtau_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -1 / s**2 * (logistic(X) * logistic(-X))


def actr_ddq1_dd_dd_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -1 / s**2 * (logistic(X) * logistic(-X)) * da_dd(
        deltati, d
    ) ** 2 + 1 / s * logistic(-X) * dda_ddd(deltati, d)


def actr_ddq1_ds_dtau_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return (tau - a) / s**3 * (logistic(X) * logistic(-X)) + 1 / s**2 * logistic(-X)


def actr_ddq1_dd_dtau_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return (logistic(X) * logistic(-X)) * 1 / s**2 * da_dd(deltati, d)


def actr_ddq1_dd_ds_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -1 / s**2 * logistic(-X) * da_dd(deltati, d) - (tau - a) / s**3 * da_dd(
        deltati, d
    ) * (logistic(X) * logistic(-X))


# omega = 0


def actr_ddq0_ds_ds_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -2 * (a - tau) / s**3 * logistic(X) - (tau - a) ** 2 / s**4 * (
        logistic(X) * logistic(-X)
    )


def actr_ddq0_dtau_dtau_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -1 / s**2 * (logistic(X) * logistic(-X))


def actr_ddq0_dd_dd_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return -1 / s**2 * (logistic(X) * logistic(-X)) * da_dd(
        deltati, d
    ) ** 2 - 1 / s * logistic(X) * dda_ddd(deltati, d)


def actr_ddq0_ds_dtau_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return (tau - a) / s**3 * (logistic(X) * logistic(-X)) - 1 / s**2 * logistic(X)


def actr_ddq0_dd_dtau_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return (logistic(X) * logistic(-X)) * 1 / s**2 * da_dd(deltati, d)


def actr_ddq0_dd_ds_sample(tau, s, d, deltati):
    a = activation(deltati, d)
    X = (a - tau) / s
    return 1 / s**2 * logistic(X) * da_dd(deltati, d) - (tau - a) / s**3 * da_dd(
        deltati, d
    ) * (logistic(X) * logistic(-X))


if __name__ == "__main__":
    # [startdoc]
    from pyrbit.mle_utils import CI_asymptotical, confidence_ellipse
    from pyrbit.actr import (
        ACTR,
        diagnostics,
        actr_observed_information_matrix,
        identify_actr_from_recall_sequence,
        gen_data,
    )

    import numpy
    import matplotlib.pyplot as plt

    SEED = None
    N = 10000

    d = 0.6
    tau = -0.7
    s = 0.25
    rng = numpy.random.default_rng(seed=SEED)

    # ==================== Simulate some data
    actr = ACTR(1, 0.5, 0.25, -0.7, buffer_size=16, seed=SEED)
    recalls, deltatis = gen_data(actr, N)

    # ================= Run Diagnostics (logistic regression)
    ax = diagnostics(
        0.6,
        deltatis,
        recalls,
        line_kws={"color": "green"},
        recall_event_kwargs={"scatter_kws": {"marker": "*", "s": 5}},
    )
    ax.legend()
    plt.tight_layout()
    plt.show()

    # ==================== Perform ML Estimation
    optim_kwargs = {"method": "L-BFGS-B", "bounds": [(0, 1), (-5, 5), (-5, 5)]}
    verbose = False
    guess = (0.2, 0.5, -1)

    # An illustration of identifying with basin_hopping. This slows down inference, as optimization is repeated 1+niter times
    def _callable(x, f, accept):
        print(x, f, accept)

    inference_results = identify_actr_from_recall_sequence(
        recalls,
        deltatis,
        optim_kwargs=optim_kwargs,
        verbose=verbose,
        guess=guess,
        basin_hopping=True,
        basin_hopping_kwargs={"niter": 3, "callback": _callable},
    )
    # see scipy.optimize doc for returned object with basinhopping
    x = inference_results.lowest_optimization_result.x

    # ================= computing Confidence Intervals and Ellipses
    J = actr_observed_information_matrix(recalls, deltatis, *x)
    # covariance matrix
    covar = numpy.linalg.inv(J)
    # Confidence intervals
    cis = CI_asymptotical(covar, x)
    # Confidence ellipses
    # draw the three ellipses, because the confidence ellipsoid will be hard to interpret.
    fig, axs = plt.subplots(nrows=1, ncols=3)
    labels = [r"$d$", r"$s$", r"$\tau$"]
    for n in range(3):
        i = (n + 1) % 3
        j = (n + 2) % 3
        if i > j:
            i, j = j, i
        _cov = numpy.array([[covar[i, i], covar[i, j]], [covar[i, j], covar[j, j]]])
        ax = confidence_ellipse((x[i], x[j]), _cov, ax=axs[n])
        ax.set_xlabel(f"{labels[i]}")
        ax.set_ylabel(f"{labels[j]}")
    plt.tight_layout()
    plt.show()



from seaborn.regression import _RegressionPlotter
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy
import copy

# code patched from seaborn v0.12.2
class _PatchedRegressionPlotter(_RegressionPlotter):
    def scatterplot(self, ax, kws):
        """Draw the data."""
        # Treat the line-based markers specially, explicitly setting larger
        # linewidth than is provided by the seaborn style defaults.
        # This would ideally be handled better in matplotlib (i.e., distinguish
        # between edgewidth for solid glyphs and linewidth for line glyphs
        # but this should do for now.
        line_markers = ["1", "2", "3", "4", "+", "x", "|", "_"]
        if self.x_estimator is None:
            if "marker" in kws and kws["marker"] in line_markers:
                lw = mpl.rcParams["lines.linewidth"]
            else:
                lw = mpl.rcParams["lines.markeredgewidth"]
            kws.setdefault("linewidths", lw)

            if not hasattr(kws["color"], "shape") or kws["color"].shape[1] < 4:
                kws.setdefault("alpha", 0.8)

            x, y = self.scatter_data
            ax.scatter(x, y, **kws)
        else:
            # TODO abstraction
            ci_kws = {"color": kws["color"]}
            if "alpha" in kws:
                ci_kws["alpha"] = kws["alpha"]
            ci_kws["linewidth"] = mpl.rcParams["lines.linewidth"] * 1.75
            kws.setdefault("s", 50)

            xs, ys, cis = self.estimate_data
            if [ci for ci in cis if ci is not None]:
                for x, ci in zip(xs, cis):
                    ax.plot([x, x], ci, **ci_kws)
            ax.scatter(xs, ys, **kws)
            self.xs = xs
            self.ys = ys
            self.cis = cis

    def fit_statsmodels(self, grid, model, **kwargs):
        """More general regression function using statsmodels objects."""
        import statsmodels.tools.sm_exceptions as sme
        import warnings

        X, y = numpy.c_[numpy.ones(len(self.x)), self.x], self.y
        grid = numpy.c_[numpy.ones(len(grid)), grid]

        def reg_func(_x, _y, store=False):
            err_classes = (sme.PerfectSeparationError,)
            try:
                with warnings.catch_warnings():
                    if hasattr(sme, "PerfectSeparationWarning"):
                        # statsmodels>=0.14.0
                        warnings.simplefilter("error", sme.PerfectSeparationWarning)
                        err_classes = (*err_classes, sme.PerfectSeparationWarning)
                    _model = model(_y, _x, **kwargs).fit()
                    if store is True:
                        self.reg_params = _model.params
                    yhat = _model.predict(grid)
            except err_classes:
                yhat = numpy.empty(len(grid))
                yhat.fill(numpy.nan)
            return yhat

        yhat = reg_func(X, y, store=True)
        if self.ci is None:
            return yhat, None
        from seaborn import algorithms as algo

        yhat_boots = algo.bootstrap(
            X, y, func=reg_func, n_boot=self.n_boot, units=self.units, seed=self.seed
        )
        return yhat, yhat_boots


def regplot(
    data=None,
    *,
    x=None,
    y=None,
    x_estimator=None,
    x_bins=None,
    x_ci="ci",
    scatter=True,
    fit_reg=True,
    ci=95,
    n_boot=1000,
    units=None,
    seed=None,
    order=1,
    logistic=False,
    lowess=False,
    robust=False,
    logx=False,
    x_partial=None,
    y_partial=None,
    truncate=True,
    dropna=True,
    x_jitter=None,
    y_jitter=None,
    label=None,
    color=None,
    marker="o",
    scatter_kws=None,
    line_kws=None,
    ax=None,
):
    """
    # regplot.xs, regplot.ys, regplot.cis --> summary data

    """
    plotter = _PatchedRegressionPlotter(
        x,
        y,
        data,
        x_estimator,
        x_bins,
        x_ci,
        scatter,
        fit_reg,
        ci,
        n_boot,
        units,
        seed,
        order,
        logistic,
        lowess,
        robust,
        logx,
        x_partial,
        y_partial,
        truncate,
        dropna,
        x_jitter,
        y_jitter,
        color,
        label,
    )

    if ax is None:
        ax = plt.gca()

    scatter_kws = {} if scatter_kws is None else copy.copy(scatter_kws)
    scatter_kws["marker"] = marker
    line_kws = {} if line_kws is None else copy.copy(line_kws)
    plotter.plot(ax, scatter_kws, line_kws)
    return ax, plotter



from pyrbit.mem_utils import MemoryModel
from pyrbit.plot_utils import regplot
from pyrbit.mle_utils import mle_sequence

import matplotlib.pyplot as plt
import numpy
import pandas
import statsmodels.formula.api as smf
import seaborn
import warnings


class ExponentialForgetting(MemoryModel):
    def __init__(self, nitems, a, b, max_a=0.5, min_a=10 ** (-6), **kwargs):
        super().__init__(nitems, **kwargs)
        if a > max_a or a < min_a:
            warnings.warn(
                f"value a = {a} is outside admissible range ({min_a} < a < {max_a}). I am clipping it"
            )
            a = numpy.clip(a, min_a, max_a)
        self.a, self.b = a, b
        self.reset()

    def reset(self, *args, a=None, b=None):
        if a is not None:
            self.a = a
        if b is not None:
            self.b = b

        # self.counters is number of presentations, not repetitions
        if len(args) == 0:
            self.counters = numpy.zeros((self.nitems, 2))
            self.counters[:, 1] = -numpy.inf
        else:
            self.counters = numpy.zeros((self.n_items, 2))
            self.counters[:, 0] = args[0]
            self.counters[:, 1] = args[1]

    def update(self, item, time, N=None):
        item = int(item)
        if N is None:
            self.counters[item, 0] += 1
        else:
            self.counters[item, 0] = N
        self.counters[item, 1] = time

    def _print_info(self):
        print(f"counters: \t {self.counters}")

    def compute_probabilities(self, time=None):
        if time is None:
            time = numpy.max(self.counters[:, 1])
        n = self.counters[:, 0]
        deltat = time - self.counters[:, 1]
        return numpy.exp(-self.a * (1 - self.b) ** (n - 1) * deltat)

    def __repr__(self):
        return f"{self.__class__.__name__}\n a = {self.a}\n b = {self.b}"


def identify_ef_from_recall_sequence(
    recall_sequence,
    deltas,
    k_vector=None,
    guess=(1e-3, 0.5),
    optim_kwargs={"method": "L-BFGS-B", "bounds": [(1e-7, 5e-1), (0, 0.99)]},
    verbose=True,
    basin_hopping=False,
    basin_hopping_kwargs=None,
):
    infer_results = mle_sequence(
        _ef_get_sequence_likelihood,
        optim_kwargs,
        guess,
        deltas,
        recall_sequence,
        k_vector,
        basin_hopping=basin_hopping,
        basin_hopping_kwargs=basin_hopping_kwargs,
    )

    if verbose:
        print(infer_results)

    return infer_results


def covar_delta_method_log_alpha(alpha, var):
    """var should be inverse of J observed information matrix"""
    grad = numpy.array([[1 / alpha / numpy.log(10), 0], [0, 1]])
    return grad.T @ var @ grad


def ef_observed_information_matrix(recall_sequence, deltas, alpha, beta, k_vector=None):
    return _ef_get_sequence_observed_information_matrix(
        recall_sequence, deltas, alpha, beta, k_vector=k_vector
    )[0]


def diagnostics(
    alpha,
    beta,
    k_repetition,
    deltas,
    recall,
    exponent_kwargs=None,
    loglogplot_kwargs=None,
):
    _exponent_kwargs = {"xbins": int(len(deltas) ** (1 / 3)), "inf_to_int": -10}
    if exponent_kwargs is not None:
        _exponent_kwargs.update(exponent_kwargs)

    _loglogplot_kwargs = {}
    if loglogplot_kwargs is not None:
        _loglogplot_kwargs.update(loglogplot_kwargs)

    exponent = [
        -alpha * (1 - beta) ** (k) * dt for (k, dt) in zip(k_repetition, deltas)
    ]

    fig, axs = plt.subplots(nrows=1, ncols=1)
    ax, regplot = plot_exponent_scatter(exponent, recall, ax=axs, **_exponent_kwargs)
    ax.legend()
    fg, ax, estim = loglogpplot(k_repetition, recall, deltas, **_loglogplot_kwargs)

    return fig, (fg, ax, estim)


def flatten(
    data,
    schedule=None,
    population_model=None,
    test_blocks=None,
    get_k_delta=False,
    replications=1,
):
    if replications != 1:
        raise NotImplementedError("not supported yet")
    if not get_k_delta:
        return data[0, 0, ...].ravel()
    if test_blocks is None:
        k, d = get_k_delta_schedule(schedule)

        k = numpy.tile(k, (replications, len(population_model), 1))
        k = k.transpose(0, 2, 1)

        d = numpy.tile(d, (replications, len(population_model), 1))
        d = d.transpose(0, 2, 1)

    else:
        k, d = get_k_delta_schedule(schedule, test_blocks=test_blocks, data=data)
        k = numpy.array(k).reshape(len(population_model), len(test_blocks)).T
        d = numpy.array(d).reshape(len(population_model), len(test_blocks)).T

    return data[0, 0, ...].ravel(), d.ravel(), k.ravel()


def get_k_delta_schedule(schedule, test_blocks=None, data=None):
    if test_blocks is None:
        return _get_k_delta_schedule_no_test_blocks(schedule)
    else:
        return _get_k_delta_schedule_test_blocks(
            schedule, test_blocks=test_blocks, data=data
        )


def _get_k_delta_schedule_no_test_blocks(schedule):
    counter = {str(i): -1 for i in range(schedule.nitems)}
    timer = {str(i): -numpy.inf for i in range(schedule.nitems)}
    k_vec = []
    delta_vec = []
    for item, time in schedule:
        k_vec.append(counter[str(item)])
        delta_vec.append(time - timer[str(item)])
        counter[str(item)] += 1
        timer[str(item)] = time
    return k_vec, delta_vec


def _get_k_delta_schedule_test_blocks(schedule, test_blocks, data):
    recall = data[0, 0, :, :].transpose(1, 0)
    k_vec = []
    delta_vec = []
    for r in recall:
        counter = {str(i): -1 for i in range(schedule.nitems)}
        timer = {str(i): -numpy.inf for i in range(schedule.nitems)}

        nr = 0
        for n, (item, time) in enumerate(schedule):
            m = int(n / (schedule.nitems * schedule.repet_trials))
            if m in test_blocks:
                k_vec.append(counter[str(item)])
                delta_vec.append(time - timer[str(item)])
                if r[nr]:
                    counter[str(item)] += 1
                    timer[str(item)] = time
                nr += 1
            else:
                counter[str(item)] += 1
                timer[str(item)] = time

    return k_vec, delta_vec


# def serialize_experiment(data, times):
#     _, block_size = data.shape
#     k_vector = []
#     deltas = []
#     for i in data:
#         k_vector += [k for k in range(-1, block_size - 1)]
#         deltas += [numpy.infty] + numpy.diff(numpy.asarray(times)).tolist()
#     data = data.reshape(
#         -1,
#     )
#     return data, numpy.asarray(k_vector), numpy.asarray(deltas)


## ============ for observed information matrix ============= ##
## p1, q1, p0, q0
def ef_p1_sample(alpha, beta, k, deltat):
    return numpy.exp(ef_q1_sample(alpha, beta, k, deltat))


def ef_p0_sample(alpha, beta, k, deltat):
    return 1 - ef_p1_sample(alpha, beta, k, deltat)


def ef_q1_sample(alpha, beta, k, deltat):
    return -alpha * (1 - beta) ** k * deltat


def ef_q0_sample(alpha, beta, k, deltat):
    return numpy.log(1 - numpy.exp(ef_q1_sample(alpha, beta, k, deltat)))


## first order derivatives
def ef_dq1_dalpha_sample(alpha, beta, k, deltat):
    return -((1 - beta) ** k) * deltat


def ef_dq1_dbeta_sample(alpha, beta, k, deltat):
    return alpha * k * (1 - beta) ** (k - 1) * deltat


def ef_dq0_dalpha_sample(alpha, beta, k, deltat):
    return (
        (1 - beta) ** k
        * deltat
        * ef_p1_sample(alpha, beta, k, deltat)
        / ef_p0_sample(alpha, beta, k, deltat)
    )


def ef_dq0_dbeta_sample(alpha, beta, k, deltat):
    return (
        -k
        * alpha
        * (1 - beta) ** (k - 1)
        * deltat
        * ef_p1_sample(alpha, beta, k, deltat)
        / ef_p0_sample(alpha, beta, k, deltat)
    )


## Second order derivatives


def ef_ddq1_dalpha_dalpha_sample(alpha, beta, k, deltat):
    # return __sym_ef_ddq1_dalpha_dalpha_sample(alpha, beta, k, deltat)
    return 0


def ef_ddq1_dalpha_dbeta_sample(alpha, beta, k, deltat):
    # return __sym_ef_ddq1_dalpha_dbeta_sample(alpha, beta, k, deltat)
    return k * (1 - beta) ** (k - 1) * deltat


def ef_ddq1_dbeta_dbeta_sample(alpha, beta, k, deltat):
    # return __sym_ef_ddq1_dbeta_dbeta_sample(alpha, beta, k, deltat)
    return -alpha * k * (k - 1) * (1 - beta) ** (k - 2) * deltat


def ef_ddq0_dalpha_dalpha_sample(alpha, beta, k, deltat):
    # return __sym_ef_ddq0_dalpha_dalpha_sample(alpha, beta, k, deltat)
    if deltat == numpy.inf:
        return 0
    with numpy.errstate(divide="raise"):
        try:
            return (
                -((1 - beta) ** (2 * k))
                * deltat**2
                * (
                    ef_p1_sample(alpha, beta, k, deltat)
                    / ef_p0_sample(alpha, beta, k, deltat) ** 2
                )
            )
        except FloatingPointError:
            return -1 / alpha**2


def ef_ddq0_dalpha_dbeta_sample(alpha, beta, k, deltat):
    # return __sym_ef_ddq0_dalpha_dbeta_sample(alpha, beta, k, deltat)
    if deltat == numpy.inf:
        return 0
    with numpy.errstate(divide="raise", invalid="raise"):
        try:
            return (
                -k
                * (1 - beta) ** (k - 1)
                * deltat
                * ef_p1_sample(alpha, beta, k, deltat)
                / ef_p0_sample(alpha, beta, k, deltat)
                + alpha
                * k
                * (1 - beta) ** (2 * k - 1)
                * deltat**2
                * ef_p1_sample(alpha, beta, k, deltat)
                / ef_p0_sample(alpha, beta, k, deltat) ** 2
            )
        except FloatingPointError:
            return 0


def ef_ddq0_dbeta_dbeta_sample(alpha, beta, k, deltat):
    # return __sym_ef_ddq0_dbeta_dbeta_sample(alpha, beta, k, deltat)
    if deltat == numpy.inf:
        return 0
    with numpy.errstate(divide="raise", invalid="raise"):
        try:
            return (
                alpha
                * k
                * (k - 1)
                * (1 - beta) ** (k - 2)
                * deltat
                * ef_p1_sample(alpha, beta, k, deltat)
                / ef_p0_sample(alpha, beta, k, deltat)
                - alpha**2
                * k**2
                * deltat**2
                * (1 - beta) ** (2 * k - 2)
                * ef_p1_sample(alpha, beta, k, deltat)
                / ef_p0_sample(alpha, beta, k, deltat) ** 2
            )
        except FloatingPointError:
            return -k * (k * beta - 1) / (1 - beta)


def _ef_log_likelihood_sample(recall, k, deltat, alpha, beta, transform):
    # rescaling value to linear
    a, b = transform(alpha, beta)
    if recall == 1:  # Warning: passing to array converts recall to float
        return -a * (1 - b) ** k * deltat
    elif recall == 0:
        with numpy.errstate(over="raise", invalid="raise", divide="raise"):
            try:
                exp = numpy.exp(-a * (1 - b) ** k * deltat)

                # exp = numpy.clip(exp, a_min=0, a_max=1 - 1e-4)
                return numpy.log(1 - exp)
            except FloatingPointError:
                return -1e6

    else:
        raise ValueError(f"Recall is not 0 or 1, but is {recall}")


def _ef_get_sequence_likelihood(
    theta,
    deltas,
    recalls,
    k_vector=None,
):
    transform = lambda a, b: (a, b)
    return __ef_get_sequence_likelihood_with_transform(
        theta,
        deltas,
        recalls,
        k_vector=k_vector,
        transform=transform,
    )


def _ef_get_sequence_observed_information_matrix(
    recall_sequence, deltas, alpha, beta, k_vector=None
):
    J_11 = 0
    J_12 = 0
    J_22 = 0
    if k_vector is None:
        k_vector = list(range(1, len(deltas)))

    for recall, delta, k in zip(recall_sequence, deltas, k_vector):
        if recall == 1:
            J_11 += ef_ddq1_dalpha_dalpha_sample(alpha, beta, k, delta)
            J_12 += ef_ddq1_dalpha_dbeta_sample(alpha, beta, k, delta)
            J_22 += ef_ddq1_dbeta_dbeta_sample(alpha, beta, k, delta)
        elif recall == 0:
            J_11 += ef_ddq0_dalpha_dalpha_sample(alpha, beta, k, delta)
            J_12 += ef_ddq0_dalpha_dbeta_sample(alpha, beta, k, delta)
            J_22 += ef_ddq0_dbeta_dbeta_sample(alpha, beta, k, delta)
        else:
            raise ValueError(f"recall is not either 1 or 0, but is {recall}")

    J = -numpy.array([[J_11, J_12], [J_12, J_22]])
    return J, len(deltas)


def __ef_get_sequence_likelihood_with_transform(
    theta,
    deltas,
    recalls,
    k_vector=None,
    transform=None,
):
    ll = 0
    alpha, beta = theta

    if k_vector is None:
        for nsched, recall in enumerate(recalls):
            dll = _ef_log_likelihood_sample(
                recall, nsched, deltas[nsched], alpha, beta, transform
            )
            ll += dll
    else:
        for n, (k, recall) in enumerate(zip(k_vector, recalls)):
            dll = _ef_log_likelihood_sample(
                recall, k, deltas[n], alpha, beta, transform
            )
            ll += dll
    return ll


#### ============================= Plot


def plot_exponent_scatter(exponent, recall, ax=None, xbins=15, inf_to_int=None):
    if inf_to_int is not None:
        exponent = numpy.nan_to_num(exponent, neginf=inf_to_int)

    if ax is None:
        fig, ax = plt.subplots(nrows=1, ncols=1)

    _, _regplot = regplot(
        x=exponent,
        y=recall,
        ax=ax,
        fit_reg=False,
        x_bins=xbins,
        label="Estimated recall probability",
    )
    _, _ = regplot(
        x=exponent,
        y=recall,
        ax=ax,
        fit_reg=False,
        label="Recall events",
        marker=".",
        color="orange",
    )
    _x = numpy.linspace(numpy.min(exponent), numpy.max(exponent), 100)
    ax.plot(_x, [numpy.exp(x) for x in _x], "-", label="exponential forgetting model")
    ax.set_xlabel(r"$\mathrm{exponent~}\omega$")
    ax.set_ylabel("Recall (events and probabilities)")
    return ax, _regplot


def loglogpplot(k_repetition, recall, deltas, x_bins=7, mode="log"):
    sequence = [(k, r, d) for k, r, d in zip(k_repetition, recall, deltas)]
    df = pandas.DataFrame(sequence)
    df.columns = ["repetition", "recall", "deltat"]
    df = df[df["repetition"] >= 0]

    if mode == "log":
        _x_bins = numpy.logspace(
            numpy.log10(numpy.min(df["deltat"])),
            numpy.log10(numpy.max(df["deltat"])),
            x_bins,
        )
    else:
        _x_bins = numpy.linspace(
            numpy.min(df["deltat"]), numpy.max(df["deltat"]), x_bins
        )

    df["discretized_deltas"] = pandas.cut(df["deltat"], bins=_x_bins, labels=False)

    df_group = df.groupby(["repetition", "discretized_deltas"]).mean()
    df_group["log_delta"] = numpy.log(df_group["deltat"])
    df_group = df_group[df_group["recall"] != 0]
    df_group["minuslogp"] = -numpy.log(df_group["recall"])
    df_group = df_group[df_group["minuslogp"] != 0]
    df_group["logminuslogp"] = numpy.log(df_group["minuslogp"])

    df_group = df_group.reset_index()

    try:
        model = smf.mixedlm(
            "logminuslogp ~ log_delta", df_group, groups=df_group["repetition"]
        )
        _fit = model.fit()
        ri = [v[0] for v in _fit.random_effects.values()]
        beta_estim = 1 - numpy.exp(numpy.mean(numpy.diff(ri)))
        alpha_estim = numpy.exp(_fit.params["Intercept"] + ri[0])
        estim = {"alpha_estim": alpha_estim, "beta_estim": beta_estim}
    except:
        estim = None

    fg = seaborn.lmplot(
        x="log_delta",
        y="logminuslogp",
        data=df_group.reset_index(),
        hue="repetition",
    )
    ax = fg.axes.flatten()[0]
    xlim = ax.get_xlim()
    ax.plot(xlim, xlim, "r-", lw=2, label="y=x")
    ax.text(numpy.mean(xlim) + 0.2, numpy.mean(xlim) - 0.2, "y=x", color="r")
    fg.set_xlabels(r"$\log(\Delta_t)$")
    fg.set_ylabels(r"$\log(-\log p)$")
    return fg, ax, estim


if __name__ == "__main__":
    # [startdoc]
    from pyrbit.mle_utils import CI_asymptotical, confidence_ellipse
    from pyrbit.ef import (
        ExponentialForgetting,
        diagnostics,
        identify_ef_from_recall_sequence,
        ef_observed_information_matrix,
        covar_delta_method_log_alpha,
    )
    from pyrbit.mem_utils import BlockBasedSchedule, experiment
    import numpy
    import matplotlib.pyplot as plt

    SEED = None
    N = 10000
    alpha = 0.001
    beta = 0.4

    # Initialize an EF memory model with 1 item
    ef = ExponentialForgetting(1, alpha, beta, seed=SEED)

    # Helper function for simulation
    rng = numpy.random.default_rng(seed=SEED)

    def simulate_arbitrary_traj(ef, k_vector, deltas):
        recall = []
        for k, d in zip(k_vector, deltas):
            ef.update(0, 0, N=k)
            recall.append(ef.query_item(0, d))
        return recall

    # ============== Simulate some data
    k_vector = rng.integers(low=0, high=10, size=N)
    deltas = rng.integers(low=1, high=5000, size=N)
    recall_probs = simulate_arbitrary_traj(ef, k_vector, deltas)
    recall = [rp[0] for rp in recall_probs]
    k_repetition = [k - 1 for k in k_vector]

    # ================ Run diagnostics
    # some options that you can set to tweak the diagnostics output
    exponent_kwargs = dict(xbins=15, inf_to_int=None)
    loglogplot_kwargs = dict(x_bins=7, mode="lin")
    # Run the diagnostics
    fig, (fg, ax, estim) = diagnostics(
        alpha,
        beta,
        k_repetition,
        deltas,
        recall,
        exponent_kwargs=exponent_kwargs,
        loglogplot_kwargs=loglogplot_kwargs,
    )
    plt.tight_layout()
    plt.show()

    # ==================== Perform ML Estimation
    # Solver parameters; this should work well and be quite fast
    optim_kwargs = {"method": "L-BFGS-B", "bounds": [(1e-5, 0.1), (0, 0.99)]}
    verbose = False
    guess = (1e-2, 0.5)
    # Pass all the observed data to the inference function. You can use basin_hopping for more precise estimates but this will take more time, see actr.py for an example. Returns the output of scipy.optimize
    inference_results = identify_ef_from_recall_sequence(
        recall_sequence=recall,
        deltas=deltas,
        k_vector=k_repetition,
        optim_kwargs=optim_kwargs,
        verbose=verbose,
        guess=guess,
        basin_hopping=False,
        basin_hopping_kwargs=None,
    )

    ## ==== computing Confidence Intervals and Ellipses

    # Get observed information matrix
    J = ef_observed_information_matrix(
        recall, deltas, *inference_results.x, k_vector=k_repetition
    )

    fig, axs = plt.subplots(nrows=1, ncols=2)

    # Compute covariance matrix:
    covar = numpy.linalg.inv(J)
    # get 95% confidence intervals
    cis = CI_asymptotical(covar, inference_results.x, critical_value=1.96)
    ax_lin = confidence_ellipse(inference_results.x, covar, ax=axs[0])

    # with log transform --- should be better
    transformed_covar = covar_delta_method_log_alpha(inference_results.x[0], covar)
    x = [numpy.log10(inference_results.x[0]), inference_results.x[1]]
    cis = CI_asymptotical(transformed_covar, x, critical_value=1.96)
    ax_log = confidence_ellipse(x, transformed_covar, ax=axs[1])
    ax_lin.set_title("CE with linear scale")
    ax_log.set_title("CE with alpha log scale")
    plt.tight_layout()
    plt.show()

    # ========== An example of inference when using a BlockBasedSchedule
    # Define a blockbasedschedule
    schedule = BlockBasedSchedule(
        1,
        5,
        [200, 200, 200, 200, 2000, 86400, 200, 2000],
        repet_trials=1,
        seed=123,
        sigma_t=None,
    )
    # Alternative way of defining a population as a list
    population_model = [
        ExponentialForgetting(1, 10 ** (-2.5), 0.75, seed=None) for i in range(200)
    ]

    data = experiment(
        population_model,
        schedule,
        test_blocks=[
            1,
            3,
            5,
            6,
            8,
        ],  # If there are learning vs recall blocks, define which are the test blocks. This changes behavior of the memory model (false recall during test does not count as an extra repetition in the models)
        replications=1,  # untested for != 1, but is not useful
    )

    # make a big one D array out of the recall data, and get the corresponding deltas and ks
    r, d, k = flatten(
        data,
        schedule=schedule,
        population_model=population_model,
        test_blocks=[1, 3, 5, 6, 8],
        get_k_delta=True,
        replications=1,
    )

    # infer model parameters
    optim_kwargs = {"method": "L-BFGS-B", "bounds": [(1e-5, 0.1), (0, 0.99)]}
    verbose = False
    guess = (1e-2, 0.5)
    # Pass all the observed data to the inference function. You can use basin_hopping for more precise estimates but this will take more time, see actr.py for an example. Returns the output of scipy.optimize
    inference_results = identify_ef_from_recall_sequence(
        recall_sequence=r,
        deltas=d,
        k_vector=k,
        optim_kwargs=optim_kwargs,
        verbose=verbose,
        guess=guess,
        basin_hopping=False,
        basin_hopping_kwargs=None,
    )
    J = ef_observed_information_matrix(r, d, *inference_results.x, k_vector=k)
    covar = numpy.linalg.inv(J)
    # get 95% confidence intervals
    transformed_covar = covar_delta_method_log_alpha(inference_results.x[0], covar)
    x = [numpy.log10(inference_results.x[0]), inference_results.x[1]]
    cis = CI_asymptotical(transformed_covar, x, critical_value=1.96)
    print(cis)



import pymc
from pyrbit.mem_utils import experiment, GaussianPopulation, BlockBasedSchedule
from pyrbit.ef import ExponentialForgetting
import functools
import matplotlib.pyplot as plt
import numpy
import arviz as az


def simulator_block_ef(
    SIGMA,
    repet_trials,
    nitems,
    pop_size,
    replications,
    intertrial_time,
    interblock_time,
    RECALL_BLOCKS,
    ALPHA,
    BETA,
):
    if repet_trials != 1:
        raise NotImplementedError
    if replications != 1:
        raise NotImplementedError

    default_population_kwargs = {
        "mu": [ALPHA, BETA],
        "sigma": SIGMA,
        "seed": None,
        "n_items": nitems,
        "population_size": pop_size,
    }
    schedule = BlockBasedSchedule(
        nitems, intertrial_time, interblock_time, repet_trials=repet_trials
    )

    population_model = GaussianPopulation(
        ExponentialForgetting,
        **default_population_kwargs,
    )

    data = experiment_test_blocks(
        population_model, schedule, test_blocks=RECALL_BLOCKS, replications=1
    )

    data = data[:, 0, ...]

    block_average_mean = numpy.mean(data, axis=0)
    block_std = numpy.std(data, axis=0)
    return numpy.mean(block_average_mean, axis=1)


def ef_simulator(
    repet_trials,
    nitems,
    pop_size,
    replications,
    intertrial_time,
    interblock_time,
    test_blocks,
    sigma_t,
    seed,
    alpha,
    beta,
    sigma,
):

    schedule = BlockBasedSchedule(
        nitems,
        intertrial_time,
        interblock_time,
        repet_trials=repet_trials,
        seed=seed,
        sigma_t=sigma_t,
    )

    population_model = GaussianPopulation(
        ExponentialForgetting,
        mu=[alpha, beta],
        sigma=sigma,
        population_size=pop_size,
        n_items=nitems,
        seed=seed,
    )

    # for model in population_model:
    #     print(model)
    # exit()

    data_A = experiment(
        population_model, schedule, test_blocks=test_blocks, replications=replications
    )
    data = data_A[0, 0, ...]
    return (
        numpy.mean(data, axis=1),  # mean
        numpy.std(data, axis=1) / numpy.sqrt(data.shape[-1]),  # standard error
        data,
    )


# def ef_simulator(schedule, population_kwargs):

#     default_population_kwargs = {"mu": [1e-2, .4], "sigma": 1e-6*numpy.eye(2), "seed": None}
#     default_population_kwargs.update(population_kwargs)

#     population_model = GaussianPopulation(ExponentialForgetting, **population_kwargs)
#     data = experiment(population_model=population_model, schedule, replications=1)
#     nblock = len(schedule.interblock_time) + 1

#     data = (
#         data[0, 0, ...]
#         .transpose(1, 0)
#         .reshape(
#             population_model.pop_size,
#             nblock,
#             schedule.nitems * schedule.repet_trials,
#         )
#     )
#     return data.mean(axis=(0, 2)).squeeze(), data.std(axis=(0, 2)).squeeze(), data


def ef_infer_abc(observed_data, sim, simulator_kwargs=None):
    sim_kwargs = {
        "epsilon": 0.01,
        "observed": observed_data,
        "distance": "gaussian",
        "sum_stat": "identity",
        "ndims_params": [0, 0],
    }
    if simulator_kwargs is not None:
        sim_kwargs.update(simulator_kwargs)

    with pymc.Model() as _model:
        a = pymc.Uniform("log10alpha", -6, -0.5)
        b = pymc.Uniform("b", 0.01, 0.99)

        s = pymc.Simulator("s", sim, params=(a, b), **sim_kwargs)

        idata = pymc.sample_smc(parallel=True, kernel="ABC")
        idata.extend(pymc.sample_posterior_predictive(idata))

    return idata


def plot_ihd_contours(idata, ax=None):
    if ax is None:
        fig, ax = plt.subplots(nrows=1, ncols=1)
    log10_alpha_posterior = numpy.concatenate(idata.posterior["log10alpha"].values)
    beta_posterior = numpy.concatenate(idata.posterior["b"].values)
    az.plot_kde(
        log10_alpha_posterior,
        values2=beta_posterior,
        ax=ax,
        hdi_probs=[0.68, 0.95],
        contourf_kwargs={"colors": ["#B0E0E6", "#87CEEB"]},
        legend=False,
    )
    return ax


if __name__ == "__main__":

    data = numpy.random.normal(loc=0, scale=1, size=1000)

    def normal_sim(rng, a, b, size=1000):
        return rng.normal(a, b, size=size)

    with pymc.Model() as example:
        a = pymc.Normal("a", mu=0, sigma=5)
        b = pymc.HalfNormal("b", sigma=1)
        s = pymc.Simulator(
            "s", normal_sim, params=(a, b), sum_stat="sort", epsilon=1, observed=data
        )

        idata = pymc.sample_smc()
        idata.extend(pymc.sample_posterior_predictive(idata))



import numpy
import scipy
import scipy.optimize as opti
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse
import pandas
import seaborn

class InvalidConfidenceEllipsisError(Exception):
    """InvalidConfidenceEllipsisError

    Raised when the confidence ellipsis can not be computed properly.
    """

    pass


def mle_sequence(
    likelihood_function,
    optimizer_kwargs,
    guess,
    *args,
    basin_hopping=False,
    basin_hopping_kwargs=None,
    **kwargs,
):
    # note: removed possibility to specify transform function
    # note: check the k_vector

    # invert sign and deal with argument shape
    ll_lambda = lambda guess, args: -likelihood_function(guess, *args)
    if not basin_hopping:
        res = opti.minimize(ll_lambda, guess, args=(args,), **optimizer_kwargs)
    else:
        if basin_hopping_kwargs is None:
            basin_hopping_kwargs = {}
        optimizer_kwargs.update({"args": (args,)})
        res = opti.basinhopping(
            ll_lambda, guess, minimizer_kwargs=optimizer_kwargs, **basin_hopping_kwargs
        )

    return res


def _confidence_ellipsis(x, cov, critical_value=5.991, **kwargs):
    # critical values can be looked up in a chisquared table with df = 2

    eigen_values, eigen_vectors = numpy.linalg.eig(cov)
    indexes = eigen_values.argsort()[::-1]
    eigen_values = eigen_values[indexes]
    eigen_vectors = eigen_vectors[:, indexes]

    ellipsis_orientation = numpy.arctan2(eigen_vectors[:, 0][1], eigen_vectors[:, 0][0])
    with numpy.errstate(invalid="raise"):
        ellipsis_large_axis = 2 * numpy.sqrt(critical_value * eigen_values[0])
        ellipsis_small_axis = 2 * numpy.sqrt(critical_value * eigen_values[1])
    return Ellipse(
        x, ellipsis_large_axis, ellipsis_small_axis, ellipsis_orientation, **kwargs
    )


def confidence_ellipse(
    inferred_parameters,
    estimated_covariance_matrix,
    confidence_levels=[0.68, 0.95],
    ax=None,
    colors=["#B0E0E6", "#87CEEB"],
    plot_kwargs={"color": "red", "marker": "D", "label": "ML estimate"},
):
    """estimated_covariance_matrix = numpy.linalg.inv(J)"""

    x = inferred_parameters

    if ax is None:
        fig, ax = plt.subplots(nrows=1, ncols=1)

    critical_values = [scipy.stats.chi2.ppf(cl, 2) for cl in confidence_levels]
    for critical_value, color, cl in zip(
        critical_values[::-1], colors[::-1], confidence_levels[::-1]
    ):
        try:
            _patch = _confidence_ellipsis(
                x,
                estimated_covariance_matrix,
                critical_value=critical_value,
                fill=True,
                facecolor=color,
                edgecolor="b",
                # label=f"Confidence level:{cl}",
            )
            ax.add_patch(_patch)
        except FloatingPointError:
            raise InvalidConfidenceEllipsisError(
                "The confidence ellipsis is invalid and can not be computed. Most likely the estimated covariance matrix is not positive"
            )

    ax.plot(*x, **plot_kwargs)

    return ax


def CI_asymptotical(invJ, params, critical_value=1.96):
    if invJ.shape[0] != len(params):
        raise ValueError(
            f"Size of params ({len(params)}) different from dimension of the covariance matrix ({invJ.shape[0]})"
        )
    cis = []
    for np, param in enumerate(params):

        with numpy.errstate(invalid="raise"):
            try:
                ci_low = param - critical_value * numpy.sqrt(invJ[np, np])
                ci_high = param + critical_value * numpy.sqrt(invJ[np, np])
            except FloatingPointError:
                ci_low = numpy.nan
                ci_high = numpy.nan

        cis.append((ci_low, ci_high))

    return cis


# https://stackoverflow.com/questions/43238173/python-convert-matrix-to-positive-semi-definite/43244194#43244194
# code by Ahmed Fasih
def nearestPD(A):
    """Find the nearest positive-definite matrix to input

    A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code [1], which
    credits [2].

    [1] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd

    [2] N.J. Higham, "Computing a nearest symmetric positive semidefinite
    matrix" (1988): https://doi.org/10.1016/0024-3795(88)90223-6
    """

    B = (A + A.T) / 2
    _, s, V = numpy.linalg.svd(B)

    H = numpy.dot(V.T, numpy.dot(numpy.diag(s), V))

    A2 = (B + H) / 2

    A3 = (A2 + A2.T) / 2

    if isPD(A3):
        return A3

    spacing = numpy.spacing(numpy.linalg.norm(A))
    # The above is different from [1]. It appears that MATLAB's `chol` Cholesky
    # decomposition will accept matrixes with exactly 0-eigenvalue, whereas
    # Numpy's will not. So where [1] uses `eps(mineig)` (where `eps` is Matlab
    # for `np.spacing`), we use the above definition. CAVEAT: our `spacing`
    # will be much larger than [1]'s `eps(mineig)`, since `mineig` is usually on
    # the order of 1e-16, and `eps(1e-16)` is on the order of 1e-34, whereas
    # `spacing` will, for Gaussian random matrixes of small dimension, be on
    # othe order of 1e-16. In practice, both ways converge, as the unit test
    # below suggests.
    I = numpy.eye(A.shape[0])
    k = 1
    while not isPD(A3):
        mineig = numpy.min(numpy.real(numpy.linalg.eigvals(A3)))
        A3 += I * (-mineig * k ** 2 + spacing)
        k += 1

    return A3


def isPD(B):
    """Returns true when input is positive-definite, via Cholesky"""
    try:
        _ = numpy.linalg.cholesky(B)
        return True
    except numpy.linalg.LinAlgError:
        return False


def compute_summary_statistics_estimation(estimated_parameters, subsample_sequence, TRUE_VALUE, ax = None,      bias_kwargs=None, std_kwargs=None,):
    if bias_kwargs is None:
        bias_kwargs = {}
    if std_kwargs is None:
        std_kwargs = {}

    n=estimated_parameters.shape[0]
    N = estimated_parameters.shape[1]
    
    TRUE_VALUE = numpy.repeat(
        numpy.asarray(TRUE_VALUE)[:, None],N , axis=1
    )

    estimated_parameters_bias = numpy.abs(
        numpy.nanmean(estimated_parameters, axis=2) - TRUE_VALUE
    )
    estimated_parameters_std = numpy.nanstd(estimated_parameters, axis=2)

    k = estimated_parameters_bias.shape[1]
    agg_data = numpy.zeros(shape=(n * k, 4))
    agg_data[:k, 0] = estimated_parameters_bias[0, :]
    agg_data[k : n * k, 0] = estimated_parameters_bias[1, :]
    agg_data[:k, 1] = estimated_parameters_std[0, :]
    agg_data[k : n * k, 1] = estimated_parameters_std[1, :]
    agg_data[:k, 2] = subsample_sequence
    agg_data[k : n * k, 2] = subsample_sequence
    agg_data[:k, 3] = 0
    agg_data[k : n * k, 3] = 1

    df = pandas.DataFrame(agg_data, columns=["|Bias|", "Std dev", "N", "parameter"])
    if n == 2:
        mapping = {"0": r"$\alpha$", "1": r"$\beta$"}
    else:
        raise NotImplementedError
    df["parameter"] = df["parameter"].map(lambda s: mapping.get(str(int(s))))
    df["N"] = df["N"].astype(int)

    if ax is None:
        return agg_data, df
    
    seaborn.barplot(
        data=df, x="N", y="|Bias|", hue="parameter", ax=ax[0], **bias_kwargs
    )
    seaborn.barplot(
        data=df, x="N", y="Std dev", hue="parameter", ax=ax[1], **std_kwargs
    )

    ax[0].set_yscale("log")
    ax[1].set_yscale("log")

    return agg_data, df, ax


from pyrbit.ef import (
    identify_ef_from_recall_sequence,
    ef_ddq0_dalpha_dalpha_sample,
    ef_ddq0_dalpha_dbeta_sample,
    ef_ddq0_dbeta_dbeta_sample,
    ef_ddq1_dalpha_dalpha_sample,
    ef_ddq1_dalpha_dbeta_sample,
    ef_ddq1_dbeta_dbeta_sample,
    ef_dq1_dalpha_sample,
    ef_dq1_dbeta_sample,
    ef_dq0_dalpha_sample,
    ef_dq0_dbeta_sample,
    ef_observed_information_matrix,
)
from pyrbit.mle_utils import nearestPD, compute_summary_statistics_estimation

import numpy
import pandas
import seaborn
from tqdm import tqdm
import json


def get_sample_observed_information(recall, delta, k, alpha, beta):
    J_11 = 0
    J_12 = 0
    J_22 = 0
    if recall == 1:
        J_11 += ef_ddq1_dalpha_dalpha_sample(alpha, beta, k, delta)
        J_12 += ef_ddq1_dalpha_dbeta_sample(alpha, beta, k, delta)
        J_22 += ef_ddq1_dbeta_dbeta_sample(alpha, beta, k, delta)
    elif recall == 0:
        J_11 += ef_ddq0_dalpha_dalpha_sample(alpha, beta, k, delta)
        J_12 += ef_ddq0_dalpha_dbeta_sample(alpha, beta, k, delta)
        J_22 += ef_ddq0_dbeta_dbeta_sample(alpha, beta, k, delta)
    else:
        raise ValueError(f"recall is not either 1 or 0, but is {recall}")

    J = -numpy.array([[J_11, J_12], [J_12, J_22]])
    return J


def get_sample_J_for_sequence(recall_seq, deltas, k_vector, alpha, beta):
    J_list = []
    recall_sequence = []
    for r, d, k in zip(recall_seq, deltas, k_vector):
        J_list.append(get_sample_observed_information(r, d, k, alpha, beta))
        recall_sequence.append(r)

    return J_list, recall_sequence


def get_cumulative_J_for_sequence(recall_seq, deltas, k_vector, alpha, beta):
    J_list = []
    recall_sequence = []
    for n, (r, d, k) in enumerate(zip(recall_seq, deltas, k_vector)):
        J_list.append(
            ef_observed_information_matrix(
                recall_seq[:n], deltas[:n], alpha, beta, k_vector=k_vector[:n]
            )
        )
        recall_sequence.append(r)

    return J_list, recall_sequence


def get_sample_score(recall, delta, k, alpha, beta):
    s_1 = 0
    s_2 = 0
    if recall == 1:
        s_1 += ef_dq1_dalpha_sample(alpha, beta, k, delta)
        s_2 += ef_dq1_dbeta_sample(alpha, beta, k, delta)
    elif recall == 0:
        s_1 += ef_dq0_dalpha_sample(alpha, beta, k, delta)
        s_2 += ef_dq0_dbeta_sample(alpha, beta, k, delta)
    else:
        raise ValueError(f"recall is not either 1 or 0, but is {recall}")

    return numpy.array([s_1, s_2]).reshape(2, 1)


def get_sample_score_variance_for_sequence(recall_seq, deltas, k_vector, alpha, beta):
    J_list = []
    recall_sequence = []
    for r, d, k in zip(recall_seq, deltas, k_vector):
        score = get_sample_score(r, d, k, alpha, beta)
        J_list.append(score @ score.T)
        recall_sequence.append(r)

    return J_list, recall_sequence


def gen_hessians(
    N,
    REPETITION,
    TRUE_VALUE,
    population_model,
    play_schedule,
    subsample_sequence,
    play_schedule_args=None,
    optim_kwargs=None,
    filename=None,
    save=True,
):
    if play_schedule_args is None:
        play_schedule_args = ()

    default_optim_kwargs = {
        "method": "L-BFGS-B",
        "bounds": [(1e-5, 0.1), (0, 0.99)],
        "guess": (1e-2, 0.4),
        "verbose": False,
    }
    if optim_kwargs is None:
        pass
    else:
        default_optim_kwargs.update(optim_kwargs)

    basin_hopping = default_optim_kwargs.pop("basin_hopping", False)
    basin_hopping_kwargs = default_optim_kwargs.pop("basin_hopping", {"niter": 3})

    n_theta = len(TRUE_VALUE)
    recall_array = numpy.full((N, REPETITION), fill_value=numpy.nan)
    observed_hessians = numpy.full((n_theta**2, N, REPETITION), fill_value=numpy.nan)
    observed_scores = numpy.full((n_theta**2, N, REPETITION), fill_value=numpy.nan)
    observed_cum_hessians = numpy.full(
        (n_theta**2, N, REPETITION), fill_value=numpy.nan
    )
    estimated_parameters = numpy.full(
        (2, len(subsample_sequence), REPETITION), fill_value=numpy.nan
    )

    guess = default_optim_kwargs.pop("guess")
    verbose = default_optim_kwargs.pop("verbose")

    for repet in tqdm(range(REPETITION)):
        sequences = play_schedule(population_model, *play_schedule_args)
        J_list, recall_sequence = get_sample_J_for_sequence(*sequences, *TRUE_VALUE)
        cumJ_list, recall_sequence = get_cumulative_J_for_sequence(
            *sequences, *TRUE_VALUE
        )
        score_list, recall_sequence2 = get_sample_score_variance_for_sequence(
            *sequences, *TRUE_VALUE
        )

        recall_array[:, repet] = recall_sequence
        observed_hessians[..., repet] = numpy.array(
            [J.reshape((n_theta**2,)) for J in J_list]
        ).transpose(1, 0)
        observed_scores[..., repet] = numpy.array(
            [score.reshape((n_theta**2,)) for score in score_list]
        ).transpose(1, 0)
        observed_cum_hessians[..., repet] = numpy.array(
            [cumJ.reshape((n_theta**2,)) for cumJ in cumJ_list]
        ).transpose(1, 0)

        for ni, i in enumerate(subsample_sequence):
            idx = int(i)
            inference_results = identify_ef_from_recall_sequence(
                *[_input[:idx] for _input in sequences],
                guess=guess,
                verbose=verbose,
                optim_kwargs=default_optim_kwargs,
                basin_hopping=basin_hopping,
                basin_hopping_kwargs=basin_hopping_kwargs,
            )
            estimated_parameters[:, ni, repet] = inference_results.x
    json_data = {
        "recall_array": recall_array.tolist(),
        "observed_hessians": observed_hessians.tolist(),
        "observed_score": observed_scores.tolist(),
        "observed_cum_hessians": observed_cum_hessians.tolist(),
        "estimated_parameters": estimated_parameters.tolist(),
    }
    if filename is None:
        return json_data, filename
    if save:
        with open(filename, "w") as _file:
            json.dump(json_data, _file)

    return json_data, filename


def compute_observed_information(
    observed_hessians,
    axs=None,
    observed_information_kwargs=None,
):
    default_observed_information_kwargs = {
        "label": "Average Sample Fisher information",
    }
    if observed_information_kwargs is not None:
        default_observed_information_kwargs.update(observed_information_kwargs)

    cum_color = default_observed_information_kwargs.pop("cum_color", "red")

    fischer_cumulative = default_observed_information_kwargs.pop("cumulative", True)

    N = observed_hessians.shape[1]
    n = int(numpy.sqrt(observed_hessians.shape[0]))

    information = numpy.zeros((N,))
    cum_inf = numpy.zeros((N,))
    mean_observed_information = numpy.mean(observed_hessians, axis=2)
    mean_observed_information = mean_observed_information.transpose(1, 0)

    for ni, i in enumerate(mean_observed_information):
        try:
            information[ni] = numpy.sqrt(numpy.linalg.det(nearestPD(i.reshape(n, n))))
        except:
            print("warning -- could not compute information properly")
            information[ni] = 0

    for ni, i in enumerate(mean_observed_information):
        try:
            cum_inf[ni] = numpy.sqrt(
                numpy.linalg.det(
                    nearestPD(
                        numpy.sum(mean_observed_information[:ni, :], axis=0).reshape(
                            n, n
                        )
                    )
                )
            )
        except:
            print("warning -- could not compute information properly")
            cum_inf[ni] = 0

    if axs is None:
        return mean_observed_information, information, cum_inf

    seaborn.regplot(
        x=list(range(1, N + 1)),
        y=information,
        fit_reg=False,
        scatter=True,
        ci=None,
        ax=axs,
        label="Sample Fisher information",
    )

    seaborn.regplot(
        x=list(range(1, N + 1)),
        y=information,
        fit_reg=False,
        scatter=True,
        ax=axs,
        **default_observed_information_kwargs,
    )

    if fischer_cumulative:
        _ax = axs.twinx()

        _ax.set_ylabel("Sequence Fisher information", color=cum_color)
        _ax.tick_params(axis="y", labelcolor=cum_color)
        _ax.set_yscale("linear")
        _ax.plot(
            list(range(1, N + 1)),
            cum_inf,
            label="Sequence Fisher information",
            color=cum_color,
        )

    axs.set_xlabel("N")
    axs.set_ylabel("Observed Information")
    axs.legend()
    _ax.legend()
    return mean_observed_information, information, cum_inf, axs


def compute_full_observed_information(
    TRUE_VALUE,
    recall_array,
    observed_hessians,
    estimated_parameters,
    subsample_sequence,
    axs=None,
    recall_kwargs=None,
    observed_information_kwargs=None,
    bias_kwargs=None,
    std_kwargs=None,
):
    """
    recall_array.shape = (REPET, N)

    """
    default_recall_kwargs = {
        "fit_reg": False,
        "ci": None,
        "label": "estimated probabilities",
    }
    if recall_kwargs is not None:
        default_recall_kwargs.update(recall_kwargs)

    n = len(TRUE_VALUE)

    if axs is None:
        mean_observed_information = numpy.mean(observed_hessians, axis=2)
        mean_observed_information2 = mean_observed_information.transpose(1, 0)
        agg_data, df = compute_summary_statistics_estimation(
            estimated_parameters, subsample_sequence, TRUE_VALUE, ax=None
        )
        return mean_observed_information, agg_data, df

    X = numpy.ones((recall_array.shape))
    X = numpy.cumsum(X, axis=1)
    seaborn.regplot(
        x=X.ravel(),
        y=recall_array.ravel(),
        scatter=True,
        fit_reg=False,
        ci=None,
        ax=axs[0],
        label="events",
    )

    seaborn.regplot(
        x=X.ravel(), y=recall_array.ravel(), ax=axs[0], **default_recall_kwargs
    )

    _ax = axs[1]
    mean_observed_information, information, cum_inf, _ax = compute_observed_information(
        observed_hessians,
        axs=_ax,
        observed_information_kwargs=observed_information_kwargs,
    )
    agg_data, df, _axs = compute_summary_statistics_estimation(
        estimated_parameters,
        subsample_sequence,
        TRUE_VALUE,
        ax=[axs[2], axs[3]],
        bias_kwargs=bias_kwargs,
        std_kwargs=std_kwargs,
    )

    axs[0].set_ylim([-0.05, 1.05])
    axs[0].set_xlabel("N")
    axs[0].set_ylabel("Recalls")
    axs[0].legend()

    return (
        mean_observed_information,
        agg_data,
        information,
        cum_inf,
    )


if __name__ == "__main__":
    # [information-start]
    from pyrbit.ef import ExponentialForgetting
    from pyrbit.information import gen_hessians, compute_full_observed_information

    import matplotlib.pyplot as plt

    ALPHA_TRUE = 1e-2
    BETA_TRUE = 4e-1
    SEED = 999
    rng = numpy.random.default_rng(seed=SEED)
    REPETITION = 1000
    SUBSAMPLE = 15  # use a subsample such that N/subsample is integer. A ratio of 10 as used here should be fine for most cases.
    N = 150
    subsample_sequence = numpy.logspace(0.5, numpy.log10(N), int(N / SUBSAMPLE))

    # IID schedule --- This is what you want to modify to evaluate a different schedule
    def play_iid_schedule(ef, N):
        k_vector = rng.integers(low=-1, high=10, size=N)
        deltas = rng.integers(low=1, high=5000, size=N)
        recall_probs = simulate_arbitrary_traj(ef, k_vector, deltas)
        recall = [rp[0] for rp in recall_probs]
        k_repetition = [k for k in k_vector]
        return recall, deltas, k_repetition

    # helpfer function
    def simulate_arbitrary_traj(ef, k_vector, deltas):
        recall = []
        for k, d in zip(k_vector, deltas):
            ef.update(0, 0, N=(k + 1))
            recall.append(ef.query_item(0, d))
        return recall

    ef = ExponentialForgetting(1, ALPHA_TRUE, BETA_TRUE, seed=SEED)
    play_schedule = play_iid_schedule
    play_schedule_args = (N,)

    optim_kwargs = {
        "method": "L-BFGS-B",
        "bounds": [(1e-5, 0.1), (0, 0.99)],
        "guess": (1e-3, 0.7),
        "verbose": False,
    }
    filename = None  # change if you want to save the data (json)
    json_data, _ = gen_hessians(
        N,
        REPETITION,
        [ALPHA_TRUE, BETA_TRUE],
        ef,
        play_schedule,
        subsample_sequence,
        play_schedule_args=play_schedule_args,
        optim_kwargs=optim_kwargs,
        filename=filename,
    )

    # reshaping the output of gen_hessians
    recall_array = numpy.asarray(json_data["recall_array"])
    observed_hessians = numpy.asarray(json_data["observed_hessians"])
    observed_cum_hessians = numpy.asarray(json_data["observed_cum_hessians"])
    estimated_parameters = numpy.asarray(json_data["estimated_parameters"])
    recall_array = recall_array.transpose(1, 0)

    # parameters for the information plots
    recall_kwargs = {
        "x_bins": 10,
    }
    observed_information_kwargs = {"x_bins": 10, "cum_color": "orange"}

    fig, axs = plt.subplots(nrows=2, ncols=2, figsize=(20, 15))

    (
        fischer_information,
        agg_data,
        cumulative_information,
        cum_inf,
    ) = compute_full_observed_information(
        [ALPHA_TRUE, BETA_TRUE],
        recall_array,
        observed_hessians,
        estimated_parameters,
        subsample_sequence,
        axs=axs.ravel(),
        recall_kwargs=recall_kwargs,
        observed_information_kwargs=observed_information_kwargs,
        bias_kwargs=None,
        std_kwargs=None,
    )

    plt.tight_layout(w_pad=2, h_pad=2)
    plt.show()



import numpy
from abc import ABC, abstractmethod


class MemoryModel:
    def __init__(self, nitems, seed=None, **kwargs):
        self.nitems = nitems
        self.rng = numpy.random.default_rng(seed=seed)

    @abstractmethod
    def update(self, item, time):
        raise NotImplementedError

    @abstractmethod
    def compute_probabilities(self, time=None):
        raise NotImplementedError

    def query_item(self, item, time):
        prob = self.compute_probabilities(time=time)[item]
        return (self.rng.random() < prob, prob)


class GaussianPopulation:
    """A Population where parameters are sampled according to a Gaussian.

    You can iterate over this object until you have used up all its pop_size members.
    """

    def __init__(
        self,
        model,
        mu,
        sigma,
        population_size=1,
        n_items=1,
        seed=None,
        **kwargs,
    ):
        self.pop_size = population_size
        self.seed = numpy.random.SeedSequence(seed)
        self.n_items = n_items
        self.kwargs = kwargs
        self.rng = numpy.random.default_rng(seed=self.seed)
        self.model = model

        self.mu = mu
        self.sigma = sigma

    def __iter__(self):
        self.counter = 0
        return self

    def __next__(self):
        if self.counter < self.pop_size:
            self.counter += 1
            theta = self.rng.multivariate_normal(self.mu, self.sigma)

            return self.model(
                self.n_items,
                *theta,
                seed=self.seed.spawn(1)[0],
                **self.kwargs,
            )
        raise StopIteration

    def __len__(self):
        return self.pop_size


def run_trials(
    memory_model, schedule, reset=True, test_blocks=None, get_time_info=False
):
    if test_blocks is None:
        return _trial(memory_model, schedule, reset=reset, get_time_info=get_time_info)
    else:
        return _trial_test_blocks(
            memory_model,
            schedule,
            test_blocks,
            reset=reset,
            get_time_info=get_time_info,
        )


def _trial(memory_model, schedule, reset=True, get_time_info=False):
    """trial applies a schedule to a memory model and gathers queries


    :param memory_model: The memory model that will be queried from
    :type memory_model: inherits from MemoryModel
    :param schedule: schedule of interrogations
    :type schedule: inherits from Schedule
    :return: (queries, memory_model), where queries = [(good recall?, recall_probability)]
    :rtype: list(tuple(boolean, float))
    """
    queries = []
    times = []
    if reset:
        memory_model.reset()
    for item, time in schedule:
        query = memory_model.query_item(item, time)
        queries.append(query)
        memory_model.update(item, time)
        times.append(time)

    if get_time_info:
        return queries, memory_model, times

    return queries, memory_model


def _trial_test_blocks(
    memory_model, schedule, test_blocks, reset=True, get_time_info=False
):
    queries = []
    times = []
    if reset:
        memory_model.reset()
    for n, (item, time) in enumerate(schedule):
        if (
            int(n / (schedule.nitems * schedule.repet_trials)) in test_blocks
        ):  # if recall block
            query = memory_model.query_item(item, time)
            queries.append(query)
            times.append(time)
            if query[0]:
                memory_model.update(item, time)
        else:  # if learning block
            memory_model.update(item, time)

    if get_time_info:
        return queries, memory_model, times
    return queries, memory_model


def experiment(
    population_model,
    schedule,
    replications=1,
    test_blocks=None,
    reset=True,
    get_trial_info=False,
):
    if test_blocks is None:
        schedule_length = len(schedule)
    else:
        schedule_length = len(test_blocks) * schedule.nitems * schedule.repet_trials

    data = numpy.zeros(
        (
            replications,
            2,
            schedule_length,
            len(population_model),
        )
    )
    trial_info = numpy.zeros(
        (
            replications,
            2,
            schedule_length,
            len(population_model),
        )
    )

    for n, memory_model in enumerate(population_model):
        for i in range(replications):
            trial_data = run_trials(
                memory_model,
                schedule,
                test_blocks=test_blocks,
                reset=True,
                get_time_info=get_trial_info,
            )
            data[i, :, :, n] = numpy.array(trial_data[0]).T
            if get_trial_info:
                trial_info[i, 0, :, n] = list(range(data.shape[2]))
                trial_info[i, 1, :, n] = trial_data[2]
    if not get_trial_info:
        return data
    return data, trial_info


def reshape_experiment(data, nitems, repet_trials, nblocks):
    _shape = data.shape
    new_shape = (_shape[0], _shape[1], nblocks, repet_trials, nitems, _shape[-1])
    return data.reshape(new_shape)


# def experiment_test_blocks(population_model, schedule, test_blocks, replications=1):
#     data = numpy.zeros(
#         (replications, schedule.nitems, 2, len(test_blocks), population_model.pop_size)
#     )

#     for n, memory_model in enumerate(population_model):
#         for i in range(replications):
#             trial_data = numpy.array(
#                 trial_test_blocks(memory_model, schedule, test_blocks)[0]
#             )
#             trial_data = trial_data.reshape(-1, schedule.nitems, 2)

#             data[i, :, :, :, n] = trial_data.transpose(1, 2, 0)
#     return data.squeeze()


class Schedule:
    """Iterate over (items, times) pairs.

    A schedule is an association between presented items and the times at which these where presented. One can iterate over a schedule to get a pair (item, time).
    """

    def __init__(self, items, times):
        """__init__ _summary_

        _extended_summary_

        :param items: item identifier
        :type items: numpy array_like
        :param times: timestamp at which item was presented to participant
        :type times: numpy array_like
        """
        self.items = items
        self.nitems = len(set(items))
        self.times = times
        self.max = numpy.array(times).squeeze().shape[0]

    def __len__(self):
        return self.max

    def __getitem__(self, item):
        return (self.items[item], self.times[item])

    def __iter__(self):
        self.counter = 0
        return self

    def __next__(self):
        if self.counter < self.max:
            ret_value = self[self.counter]
            self.counter += 1
            return ret_value
        else:
            raise StopIteration

    def __repr__(self):
        return f"items: {self.items}\ntimes: {self.times}"


class BlockBasedSchedule(Schedule):
    def __init__(
        self,
        nitems,
        intertrial_time,
        interblock_time,
        repet_trials=1,
        seed=None,
        sigma_t=None,
    ):
        self.seed = seed
        self.rng = numpy.random.default_rng(self.seed)
        self.sigma_t = sigma_t

        self.nitems = nitems
        self.intertrial_time = intertrial_time
        self.interblock_time = interblock_time
        self.repet_trials = repet_trials if repet_trials is not None else 1

        self.items, self.times, self.blocks = self.regen_schedule()
        self.max = numpy.array(self.times).squeeze().shape[0]

    def regen_schedule(
        self, intertrial_time=None, interblock_time=None, repet_trials=None
    ):
        self.intertrial_time = (
            self.intertrial_time if intertrial_time is None else intertrial_time
        )
        self.interblock_time = (
            self.interblock_time if interblock_time is None else interblock_time
        )
        self.repet_trials = self.repet_trials if repet_trials is None else repet_trials

        items = []
        times = []
        blocks = []
        t = 0
        for nb, ibt in enumerate((self.interblock_time + [0])):
            for r in range(self.repet_trials):
                if self.sigma_t is not None:
                    _rdm = self.rng.lognormal(sigma=self.sigma_t, size=(self.nitems,))
                else:
                    _rdm = numpy.zeros((self.nitems,))
                for i, r in zip(range(self.nitems), _rdm):
                    items.append(i)
                    times.append(t)

                    t += self.intertrial_time + r
                    blocks.append(nb)
            t += ibt

        return items, times, blocks

    def print_schedule(self):
        print("item, time, block")
        for i, t, b in zip(self.items, self.times, self.blocks):
            print(i, t, b)


if __name__ == "__main__":
    # [schedule]
    from pyrbit.mem_utils import Schedule, BlockBasedSchedule

    items = [0, 1, 0, 1, 0, 1, 0, 0]
    times = [0, 100, 126, 200, 252, 500, 4844, 5877]

    # Building a schedule
    schedule = Schedule(items, times)
    # You can iterate over a schedule
    for item, time in schedule:
        print(item, time)

    # [block-schedule]
    # You can create a block based schedule, where you specify a constant intertrial time (which includes execution time), interblock times (this also implicitly specifies the number of blocks), and whether items are repeated. You can also add some randomness by adding a random time for each item, drawn from a lognormal distribution with scale sigma_t.
    schedule = BlockBasedSchedule(
        5, 5, [10, 20, 30, 30], repet_trials=2, seed=123, sigma_t=1
    )
    # you can print the schedule, which also shows block identifiers
    schedule.print_schedule()

    # [trial]
    # You can apply a schedule to a memory model and gather queries using run_trials
    from pyrbit.ef import ExponentialForgetting
    from pyrbit.mem_utils import run_trials

    ef = ExponentialForgetting(5, 0.01, 0.4, seed=123)
    queries, ef = run_trials(ef, schedule, reset=True)
    print(queries)
    print(ef.counters)

    # If you have learning blocks distinct from test blocks, you can use the test_blocks argument. This assumes that in a learning block the memory model is always updated, but not queried, while in a testing block, the memory model is always queried, and updated when that query is correct.
    queries, ef = run_trials(ef, schedule, test_blocks=[1, 3], reset=True)
    print(queries)
    print(ef.counters)

    # [population]
    # You can sample from a Gaussian population of memory models. Each memory model will have a different seed that spawns from the population's seed.
    population_model = GaussianPopulation(
        ExponentialForgetting,
        mu=[0.01, 0.4],
        sigma=1e-3 * numpy.array([[0.1, 0], [0, 1]]),
        population_size=3,
        n_items=5,
        seed=123,
    )
    # you can iterate over a population
    for p in population_model:
        print(p)

    # [experiment]
    # You can perform an experiment, by having members of the population perform trials. The data has shape (replication, 2, schedule_length, population_size); The dimension 2 is for (recall, recall_probability). Note that you can use a list of memory models rather than population objects.
    data = experiment(population_model, schedule, replications=4)

    population_model = GaussianPopulation(
        ExponentialForgetting,
        mu=[0.01, 0.4],
        sigma=1e-7 * numpy.array([[0.1, 0], [0, 1]]),
        population_size=4,
        n_items=5,
        seed=123,
    )
    data = experiment(population_model, schedule, replications=1)
    data, trials = experiment(
        population_model,
        schedule,
        replications=1,
        test_blocks=[1, 3, 4],
        get_trial_info=True,
    )
    # You can also reshape the experiment data to have shape (replication, 2, nblocks, repet_trials, nitems, population_size)
    rdata = reshape_experiment(data, 5, 2, 3)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyrbit']

package_data = \
{'': ['*']}

install_requires = \
['Sphinx<=7.0',
 'arviz>=0.15.1,<0.16.0',
 'matplotlib>=3.7.1,<4.0.0',
 'numpy>=1.24.2,<2.0.0',
 'pymc>=5.6.0,<6.0.0',
 'scipy>=1.10.1,<2.0.0',
 'seaborn>=0.12.2,<0.13.0',
 'sphinx-rtd-theme>=1.3.0,<2.0.0',
 'statsmodels>=0.14.0,<0.15.0',
 'sympy>=1.12,<2.0',
 'tqdm>=4.65.0,<5.0.0']

setup_kwargs = {
    'name': 'pyrbit',
    'version': '0.1.0',
    'description': 'Utilities for parametric modeling of Recall based interaction techniques',
    'long_description': None,
    'author': 'jgori',
    'author_email': 'juliengori@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



from pyrbit.mem_utils import (
    experiment,
    GaussianPopulation,
    BlockBasedSchedule,
    reshape_experiment,
)
from pyrbit.ef import ExponentialForgetting

import copy
import numpy
import pandas
import scipy
from tqdm import tqdm


def run_comparative_experiment(
    population_model_one,
    population_model_two,
    schedule,
    test_blocks=None,
    replications=1,
):
    memory_models_A = [p for p in population_model_one]
    memory_models_B = [p for p in population_model_two]

    data_A = experiment(
        memory_models_A, schedule, test_blocks=test_blocks, replications=replications
    )
    data_B = experiment(
        memory_models_B, schedule, test_blocks=test_blocks, replications=replications
    )
    return data_A, data_B


def diff_eval_block_percentages(
    population_model_one,
    population_model_two,
    schedule,
    test_blocks=None,
    output_df=True,
):
    # [run-comparative-experiment-start]
    data_A, data_B = run_comparative_experiment(
        population_model_one,
        population_model_two,
        schedule,
        test_blocks=test_blocks,
        replications=1,
    )
    # [run-comparative-experiment-after]
    if test_blocks is None:
        nblocks = len(set(schedule.blocks))
    else:
        nblocks = len(test_blocks)

    recall_events_A = reshape_experiment(
        data_A, schedule.nitems, schedule.repet_trials, nblocks
    )[:, 0, ...]
    recall_events_B = reshape_experiment(
        data_B, schedule.nitems, schedule.repet_trials, nblocks
    )[:, 0, ...]
    mean_recall_A = numpy.mean(recall_events_A, axis=(2, 3))
    mean_recall_B = numpy.mean(recall_events_B, axis=(2, 3))

    if not output_df:
        return mean_recall_A, mean_recall_B

    iter_mean_recall_A = mean_recall_A[0, ...]
    iter_mean_recall_B = mean_recall_B[0, ...]
    block1 = numpy.array(
        [i for i in range(1, nblocks + 1) for p in population_model_one]
    )
    block2 = numpy.array(
        [i for i in range(1, nblocks + 1) for p in population_model_two]
    )

    df = pandas.DataFrame(
        {
            "block recall %": numpy.concatenate(
                (iter_mean_recall_A.ravel(), iter_mean_recall_B.ravel()), axis=0
            ),
            "block": numpy.concatenate((block1, block2), axis=0),
            "Condition": numpy.concatenate(
                (
                    numpy.full(block1.shape, fill_value="A"),
                    numpy.full(block1.shape, fill_value="B"),
                ),
                axis=0,
            ),
        }
    )

    return df, mean_recall_A, mean_recall_B


def get_p_value_diff(
    population_model_one,
    population_model_two,
    schedule,
    combine_pvalues="fisher",
    test_blocks=None,
    **kwargs
):
    df, mr_A, mr_B = diff_eval_block_percentages(
        population_model_one,
        population_model_two,
        schedule,
        test_blocks=test_blocks,
        output_df=True,
    )

    # import seaborn
    # import matplotlib.pyplot as plt

    # fix, ax = plt.subplots(nrows=1, ncols=1)
    # seaborn.barplot(
    #     data=df,
    #     x="block",
    #     y="block recall %",
    #     hue="Condition",
    #     errorbar="se",
    #     ax=ax,
    # )
    # plt.show()

    p = numpy.zeros(
        len(combine_pvalues),
    )
    for n, combine_pvalue in enumerate(combine_pvalues):
        if combine_pvalue in ["fisher", "stouffer"]:
            # pvalues = scipy.stats.ttest_rel(mr_A[0, :], mr_B[0, :], axis=0).pvalue[1:]
            pvalues = scipy.stats.ttest_rel(mr_A[0, ...], mr_B[0, ...], axis=1).pvalue[
                1:
            ]
            p[n] = scipy.stats.combine_pvalues(pvalues, method=combine_pvalue).pvalue

        else:
            p[n] = combine_pvalue(mr_A[0, ...], mr_B[0, ...], **kwargs)

    return p


def get_p_values_frequency(
    populations_one,
    populations_two,
    schedule,
    combine_pvalues="fisher",
    test_blocks=None,
    significance_level=0.05,
):
    if not hasattr(combine_pvalues, "__iter__"):
        combine_pvalues = [combine_pvalues]

    p_value_container = numpy.zeros((len(populations_one), len(combine_pvalues)))
    for n, (population_model_one, population_model_two) in tqdm(
        enumerate(zip(populations_one, populations_two))
    ):
        p_value_container[n, :] = get_p_value_diff(
            population_model_one,
            population_model_two,
            schedule,
            combine_pvalues=combine_pvalues,
            test_blocks=test_blocks,
        )

    positives = numpy.sum(p_value_container < significance_level, axis=0)
    positive_rate = positives / len(p_value_container)
    negative_rate = 1 - positive_rate
    return positive_rate, negative_rate, p_value_container


if __name__ == "__main__":
    from pyrbit.mem_utils import (
        BlockBasedSchedule,
        experiment,
        GaussianPopulation,
        Schedule,
    )
    from pyrbit.ef import (
        get_k_delta_schedule,
        ExponentialForgetting,
        diagnostics,
        identify_ef_from_recall_sequence,
    )

    schedule = BlockBasedSchedule(
        1,
        0,
        [200, 200, 200, 200, 2000, 86400, 200, 2000],
        repet_trials=1,
        seed=123,
        sigma_t=None,
    )

    population_model = [
        ExponentialForgetting(1, 10 ** (-2.5), 0.75, seed=None) for i in range(200)
    ]

    data = experiment(
        population_model,
        schedule,
        test_blocks=[1, 3, 5, 6, 8],
        replications=1,
    )

    # [diff-block-percentage]

    # example with two equivalent population groups and RBITs
    population_model_one = GaussianPopulation(
        ExponentialForgetting,
        mu=[10 ** (-2.5), 0.75],
        sigma=1e-7 * numpy.array([[0.01, 0], [0, 1]]),
        population_size=24 * 16,
        n_items=1,
        seed=None,
    )

    population_model_two = GaussianPopulation(
        ExponentialForgetting,
        mu=[10 ** (-2.5), 0.75],
        sigma=1e-7 * numpy.array([[0.01, 0], [0, 1]]),
        population_size=24 * 16,
        n_items=1,
        seed=None,
    )

    # L1   R1   L2   R2    L3    R3    R4   L4    R5
    #   200  1000  200   200  2000 86400 200  2000
    schedule = BlockBasedSchedule(
        1,
        15,
        [200, 200, 1000, 200, 2000, 86400, 200, 2000],
        repet_trials=1,
        seed=123,
        sigma_t=1,
    )

    # get mean recall rates, as two arrays and optionnally as a large dataframe
    df, mA, mB = diff_eval_block_percentages(
        population_model_one,
        population_model_two,
        schedule,
        test_blocks=[1, 3, 5, 6, 8],
        output_df=True,
    )

    # plot the evaluation
    import seaborn
    import matplotlib.pyplot as plt

    fix, ax = plt.subplots(nrows=1, ncols=1)
    seaborn.barplot(
        data=df,
        x="block",
        y="block recall %",
        hue="Condition",
        errorbar="se",
        ax=ax,
    )
    plt.show()

    # [power analysis]
    # Empirical Statistical power analysis
    REPETITIONS = 100

    # Case 1: Equal RBITs, should find statistical significance alpha=0.05 at 5%
    population_model_one = [
        GaussianPopulation(
            ExponentialForgetting,
            mu=[10 ** (-2), 0.75],
            sigma=1e-7 * numpy.array([[0.01, 0], [0, 1]]),
            population_size=24,
            n_items=1,
            seed=None,
        )
        for i in range(REPETITIONS)
    ]
    population_model_two = [
        GaussianPopulation(
            ExponentialForgetting,
            mu=[10 ** (-2), 0.75],
            sigma=1e-7 * numpy.array([[0.01, 0], [0, 1]]),
            population_size=24,
            n_items=1,
            seed=None,
        )
        for i in range(REPETITIONS)
    ]

    # Define here custom methods to compute p values from mean recalls
    # compute the average of p values
    def mean_pvalue(mr_A, mr_B, **kwargs):
        pvalues = scipy.stats.ttest_rel(mr_A, mr_B, axis=1).pvalue[1:]
        return numpy.mean(pvalues)

    # Bonferonni-like correction
    def bonf(mr_A, mr_B, **kwargs):
        pvalues = scipy.stats.ttest_rel(mr_A, mr_B, axis=1).pvalue[1:]
        return numpy.min(pvalues) * len(pvalues)

    # Fisher and Stouffer methods can be called by their names to trigger their scipy implementation
    pos_one, neg_one, p_container = get_p_values_frequency(
        population_model_one,
        population_model_two,
        schedule,
        combine_pvalues=["stouffer", "fisher", bonf, mean_pvalue],
        test_blocks=None,
        significance_level=0.05,
    )
    # Type 1 errors: we reject the null that there is no difference in RBITs but there was actually none
    type1error = pos_one

    # Case 2: Unequal RBITs (alpha = 10**-2.1 versus alpha = 10**-2). The rate at which the test finds significant differences is an estimate of its power
    population_model_one = [
        GaussianPopulation(
            ExponentialForgetting,
            mu=[10 ** (-2.3), 0.75],
            sigma=1e-5 * numpy.array([[0.01, 0], [0, 1]]),
            population_size=24,
            n_items=1,
            seed=None,
        )
        for i in range(REPETITIONS)
    ]
    population_model_two = [
        GaussianPopulation(
            ExponentialForgetting,
            mu=[10 ** (-2), 0.75],
            sigma=1e-5 * numpy.array([[0.01, 0], [0, 1]]),
            population_size=24,
            n_items=1,
            seed=None,
        )
        for i in range(REPETITIONS)
    ]

    pos_two, neg_two, p_container = get_p_values_frequency(
        population_model_one,
        population_model_two,
        schedule,
        combine_pvalues=["stouffer", "fisher", bonf, mean_pvalue],
        test_blocks=None,
        significance_level=0.05,
    )

    # Type 2 errors: we failed to reject the null hypothesis that there is no difference between RBITs (even though it was False)
    type2error = neg_two
